/***********************************************************************/
/*                                                                     */
/* RIGI Utilities --- Copyright (C) 1996-1999                          */
/*                    The Rigi Group, University of Victoria           */
/*                                                                     */
/* see the file license.terms for usage and copying policies           */
/*                                                                     */
/* Author: Jim McDaniel                                                */
/*                                                                     */
/***********************************************************************/

#include <sys/types.h>
#include <sys/stat.h>
#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <string.h>

#include "uniqstr.h"
#include "htmltmpl.h"

#define HIBIT	 '\200'
#define RAB	 ('>' ^ HIBIT)
#define LAB	 ('<' ^ HIBIT)
#define HDR	 "header"
#define FTR	 "footer"
#define LCB	 '{'
#define RCB	 '}'
#define NAME_END '^'
static char *FmtAnchor = "a href=";
static char *FmtBookmark = "a name=";
static char *FmtLocalTag = "#%d";
static char *FmtEndTag = "/a";
static char *FmtPre = "<pre>";
static char *FmtEndPre = "</pre>";
static char *Fmtbr = "<br>";
static char *Fmtnbsp = "&nbsp;";
static char *hdr = "<html><head><body\240bgcolor=#ffffff>";
static char *ftr = "</body></html>";
static char *FmtPrev = "<a\240href=%s>Up</a>\n\n";
static char *FmtNext = "\n<a\240href=%s>Down</a>";
static char *NameDelimiters = "\t !\"#%&'()*+,-./:;<=>?@[\\]`{|}~";

static int HtIsName(const char c) {
  if (c == '$' || c == '_' || (c >= 'a' && c <= 'z')
      || (c >= 'A' && c <= 'Z') || (c >= '0' && c <= '9')) return 1;
  return 0;
}

static const char *HtStrstr(const char *s1, const char *s2) {
  const char *p = strstr(s1, s2);
  int i = strlen(s2);

  /* search for a substring that is preceded and followed
     by non-alphabetic characters. */
  while (p != NULL &&
	 ((p > s1 && *(p - 1) < HIBIT && strchr(NameDelimiters, *(p - 1)) == NULL)
	  || (*(p + i) && *(p + i) < HIBIT && strchr(NameDelimiters, *(p + i)) == NULL)))
    p = strstr(p + i, s2);
  return p;
}

static const char *HtMatch(char *srcLine, const UniqueString& match, size_t *len) {
  char *pToken, *pMatch;
  const char* pSrc;

  /* This routine does a match on srcLine given a match
     string consisting of a sequence of names separated
     by special characters.  The routine starts at the left-
     most name specifier and moves to the right-most until
     either the name is exhausted or a match is found. */

  pMatch = strdup(match);

  /* check where the name ends (at the first '^'), and terminate the anchor
     string there */
  if ((pToken = strchr(pMatch, NAME_END)) != NULL)
    *pToken = '\0';

  pToken = strtok(pMatch, NameDelimiters);
  while (pToken != NULL) {

    /* Look for a match on token in the match string. */
    if ((pSrc = HtStrstr(srcLine, pToken)) != NULL) {
      *len = strlen(pToken);
      free(pMatch);
      return pSrc;
    }

    /* Get the next token from the match string */
    pToken = strtok(NULL, NameDelimiters);
  }
  free(pMatch);
  return NULL;
}

int HtLoadTemplate(const char *tFile) {
  char *buff, *p, *endTmplate;
  int inHdrOrFtr = 0;
  FILE *ifp;
  struct stat tInfo;
  long buffSize = 0;

  if (stat(tFile, &tInfo) == -1) {
    fprintf(stderr, "Template file (%s) not found\n", tFile);
    return 0;
  }
  if (!tInfo.st_size) return 1;
  ifp = fopen(tFile, "r");
  if (ifp == NULL) {
    fprintf(stderr, "Error opening template file %s\n", tFile);
    return 0;
  }

  /* Construct the buffer for the file contents */
  buff = (char *)malloc((int)tInfo.st_size);
  buffSize = fread(buff, 1, (int)tInfo.st_size, ifp);
  if (ferror(ifp)) {
    fclose(ifp);
    fprintf(stderr, "Error reading template file %s\n", tFile);
    return 0;
  }
  buff[buffSize] = 0;
  endTmplate = &buff[buffSize];

  p = buff + strspn(buff, " \t\n\r");

  /* Read and parse the template */
  while (p && p < endTmplate) {

    if (!inHdrOrFtr) {
      if (!strncmp(p, HDR, strlen(HDR))) {
	p += strlen(HDR);
	p += strspn(p, " \t=");
	if (*p == LCB) {
	  p++;
	  inHdrOrFtr = 1;
	  hdr = p;
	}
      } else if (!strncmp(p, FTR, strlen(FTR))) {
	p += strlen(FTR);
	p += strspn(p, " \t\r\n=");
	if (*p == LCB) {
	  p++;
	  inHdrOrFtr = 1;
	  ftr = p;
	}
      }
    }
    if (inHdrOrFtr) {

      /* Look for closing brace marking the end of header or footer */
      while ((p = strchr(p, RCB)) != NULL && *(p - 1) == '\\') {
	memmove(p - 1, p, strlen(p) + 1);
	endTmplate--;
      }
      if (p != NULL) *p++ = 0;
      inHdrOrFtr = 0;
    } else {

      /* Skip to the first non-space character on the next line */
      p += strcspn(p, "\n\r");
      p += strspn(p, " \t\n\r");
    }
  }
  return 1;
}


void HtInsertAnchor(char *srcLine, const UniqueString& anchor,
		    const char *file, const int line) {
  char *p;
  static char lineStr[20];
  int i, j = 0, k, m;
  size_t anchorLen;

  if (srcLine == NULL || *srcLine == 0 || file == NULL) return;

	/* Perform a match of the anchor on the srcLine */
  if ((p = srcLine + (HtMatch(srcLine, anchor, &anchorLen) - srcLine)) == NULL) return;

  /* <A HREF=file[#line]> */
  k = strlen(FmtAnchor);
  m = strlen(file);
  i = k + m + 2;
  if (line > 1) {
    sprintf(lineStr, FmtLocalTag, line);
    i += j = strlen(lineStr);
  }
  memmove(p + i, p, strlen(p) + 1);
  *p++ = LAB;
  memcpy(p, FmtAnchor, k);
  for (i = 0; i < k; i++) *p++ ^= HIBIT;
  memcpy(p, file, m);
  for (i = 0; i < m; i++) *p++ ^= HIBIT;
  if (line > 1) memcpy(p, lineStr, j);
  for (i = 0; i < j; i++) *p++ ^= HIBIT;
  *p++ = RAB;
  for (i = 0; i < (int)anchorLen; i++) *p++ ^= HIBIT;

  /* </A> */
  k = strlen(FmtEndTag);
  i = k + 2;
  memmove(p + i, p, strlen(p) + 1);
  *p++ = LAB;
  memcpy(p, FmtEndTag, k);
  for (i = 0; i < k; i++) *p++ ^= HIBIT;
  *p = RAB;
  return;
}

void HtInsertBookmark(char *srcLine, const int bookmarkLine) {
  char *p = srcLine;
  static char bookmark[20];
  int i, j, k, m;

  if (srcLine == NULL || bookmarkLine <= 1) return;

  /* <A NAME=bookmark> */
  k = strlen(FmtBookmark);
  sprintf(bookmark, "%d", bookmarkLine);
  j = strlen(bookmark);
  m = strlen(FmtEndTag);
  i = k + j + 2 + m + 2;
  memmove(p + i, p, strlen(p) + 1);
  *p++ = LAB;
  memcpy(p, FmtBookmark, k);
  for (i = 0; i < k; i++) *p++ ^= HIBIT;
  memcpy(p, bookmark, j);
  for (i = 0; i < j; i++) *p++ ^= HIBIT;
  *p++ = RAB;

  /* </A> */
  *p++ = LAB;
  memcpy(p, FmtEndTag, m);
  for (i = 0; i < m; i++) *p++ ^= HIBIT;
  *p = RAB;
  return;
}

char *HtFmtLine(char *printLine, const int tabSize, const int isPre) {
  int i, j;
  char *p, *p1;

  /* Expand tabs */
  if (tabSize) {
    p = strchr(printLine, '\t');
    while (p != NULL) {
      p1 = printLine + ((p - printLine)/tabSize + 1) * tabSize;
      i = strlen(p);
      if (p1 + i - printLine > MAXRECSIZE) return printLine;
      memmove(p1, p + 1, i);
      memset(p, ' ', p1 - p);
      p = strchr(p1, '\t');
    }
  }

  /* Substitute spaces with nonbreaking spaces */
  if (!isPre) {
    p = strchr(printLine, ' ');
    while (p != NULL) {
      i = strlen(p);
      j = strlen(Fmtnbsp);
      if (j + strlen(printLine) - 1 > MAXRECSIZE)
	return printLine;
      memmove(p + j, p + 1, i);
      memcpy(p, Fmtnbsp, j);
      p = strchr(p + j, ' ');
    }

    /* Put a <br>/n at the end of the line */
    j = strlen(Fmtbr);
    i = strlen(printLine);
    if (i && printLine[i - 1] == '\n') printLine[--i] = 0;
    if (j + i + 1> MAXRECSIZE) return printLine;
    strcat(printLine, Fmtbr);
    printLine[i + j] = '\n';
    printLine[i + j + 1] = 0;
  }

  /* Turn all high bits off */
  j = strlen(printLine);
  p = printLine;
  for (i = 0; i < j; i++) *p++ &= (~HIBIT);

  return printLine;
}

char *HtGetFileHdr(char *printLine, const char *prevFile, const int isPre) {
  int i = strlen(hdr) + (isPre? strlen(FmtPre) : 0);
  int j = (prevFile == NULL)? 0 : strlen(prevFile);
  *printLine = 0;
  if (i + strlen(FmtPrev) + j >= MAXRECSIZE)
    return printLine;
  strcpy(printLine, hdr);
  if (isPre) strcat(printLine, FmtPre);
  if (j) sprintf(&printLine[i], FmtPrev, prevFile);
  return printLine;
}

char *HtGetFileFtr(char *printLine, const char *nextFile, const int isPre) {
  int j = (nextFile == NULL)? 0 : strlen(nextFile);
  *printLine = 0;
  if (strlen(ftr) + (isPre? strlen(FmtEndPre) : 0)
      + strlen(FmtNext) + j >= MAXRECSIZE)
    return printLine;
  if (j) sprintf(printLine, FmtNext, nextFile);
  if (isPre) strcat(printLine, FmtEndPre);
  strcat(printLine, ftr);
  return printLine;
}
