/***********************************************************************/
/*                                                                     */
/* RIGI Utilities --- Copyright (C) 1996-1999                          */
/*                    The Rigi Group, University of Victoria           */
/*                                                                     */
/* see the file license.terms for usage and copying policies           */
/*                                                                     */
/* Author: Jim McDaniel                                                */
/*                                                                     */
/***********************************************************************/
/* rsf2html.cc (html'ize source given a 4-tuple rsf file)

   Uses RSF file to place HTML tags in source.	The -a
   specifies a list of key words used to define anchors.
   The -b specifies a list of key words used to define
   bookmarks.  If -k maxlines is specified, the source
   files are segmented at the bookmarks and no segment
   will exceed maxLines.  If -t template is specified,
   the template is used to construct the HTML'ized source
   file.  The program writes out a new version of the RSF
   file and the source to the current working directory.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "rsffile.h"
#include "htmltmpl.h"
#include "rsf2html.h"

#define MAXFILESIZE 512

int maxLines = 0;
int tabSize = 0;

/* Miscellaneous global work variables */
static char rec[MAXRECSIZE + 1];
static char iFile[MAXFILESIZE + 1], oFile[MAXFILESIZE + 1];
static char tmpFile0[MAXFILESIZE + 1];
static FILE *ifp = NULL, *ofp = NULL;
static int iLine = 0, lineCnt = 0;

class BookmarkLess {
  typedef HtmlRsfTuple *pHtmlRsfTuple;
public:
  bool operator()(const pHtmlRsfTuple& x, const pHtmlRsfTuple& y) const {
    return x->rsf[(x->rsf[0] == RsfTuple::typeString) ? 1 : 2]
	 < y->rsf[(y->rsf[0] == RsfTuple::typeString) ? 1 : 2];
  }
};

typedef ptrSet<HtmlRsfTuple, BookmarkLess> bookmarkSet;

static int WriteLine(char *printLine, bool preformatted) {
  if (fputs(HtFmtLine(printLine, tabSize, preformatted), ofp) == EOF) {
    fprintf(stderr, "Write error on %s\n", oFile);
    return 0;
  }
  lineCnt++;
  return 1;
}

static void Substitute(char *line, const char c, const char *substr) {
  char *p;
  int i = strlen(substr);
  while ((p = strchr(line, c)) != NULL) {
    memmove(p + i - 1, p, strlen(p) + 1);
    memcpy(p, substr, i);
    p += i;
  }
}

static int ReadNextLine() {
  if (fgets(rec, MAXRECSIZE, ifp) == NULL) {
    if (ferror(ifp)) {
      fprintf(stderr, "Read error on %s\n", iFile);
      return 0;
    }
    rec[0] = '\0';
    return EOF;
  }
  Substitute(rec, '<', "&lt;");
  return 1;
}

static char *MkOFileName(const char *iName, char *oName, int line) {
  sprintf(oName, "%s_%04d.html", iName, line ? line : 1);

  /* replace all forward/backward slashes and colons by underlines */
  char *p;

  while (((p = strchr(oName, '/')) != NULL) ||
	 ((p = strchr(oName, '\\')) != NULL) ||
	 ((p = strchr(oName, ':')) != NULL))
    *p = '_';

  return oName;
}

static int NextOFile(int prevLink, int nextLink, bool preformatted) {
  char tmpFile1[MAXFILESIZE + 1];
  char buff[MAXRECSIZE + 1];

  if (lineCnt == 1) return 1;
  MkOFileName(iFile, tmpFile0, iLine);
  if (ofp != NULL) {
    HtGetFileFtr(buff, (nextLink? tmpFile0 : NULL), preformatted);
    if (!WriteLine(buff, preformatted)) return 0;
    fclose(ofp);
  }
  strcpy(tmpFile1, oFile);
  strcpy(oFile, tmpFile0);
  if ((ofp = fopen(oFile, "w")) == NULL) {
    fprintf(stderr, "Open error on %s\n", oFile);
    return 0;
  }
  lineCnt = 0;
  HtGetFileHdr(buff, (prevLink? tmpFile1 : NULL), preformatted);
  return WriteLine(buff, preformatted);
}

static int NextIFile(const char *rsfFile) {
  const char *p0, *p1;
  int i = 0;

  if (ifp != NULL) fclose(ifp);
  iFile[0] = 0;

  /* copy the rsfFile to iFile while omitting
     doubled backslashes (\\) */
  for (p1 = rsfFile; ; ) {
    p0 = p1;
    if ((p1 = strchr(p0, '\\')) == NULL) break;
    strncpy(&iFile[i], p0, ++p1 - p0);
    i += p1 - p0;
    while (*p1 == '\\') p1++;
  }
  strcpy(&iFile[i], p0);
  if ((ifp = fopen(iFile, "r")) == NULL) {
    fprintf(stderr, "Open error on %s\n", iFile);
    /* ignore the error to keep htmlrsf running // return 0; */
    ifp = fopen("/dev/zero", "r");
  }
  iLine = 1;

  /* Always read one line ahead */
  return ReadNextLine();
}

static int CopyToLine(const int rsfLine, bool preformatted) {
  int i;
  if (ifp == NULL) return 1;
  if ((rsfLine == EOF) && (rec[0] == '\0')) return 1;

  /* If rsfLine is EOF, then copy to end of input file.
     Otherwise copy until rsfLine is reached */
  while (rsfLine == EOF || iLine < rsfLine) {
    iLine++;
    if (WriteLine(rec, preformatted) == 0) return 0;

    /* Read ahead one line */
    if ((i = ReadNextLine()) == 0) return 0;
    if (i == EOF) {

      if (rsfLine != EOF) {
	fprintf(stderr, "Unexpected end of file on %s\n", iFile);
	/* ignore the error to keep htmlrsf running // return 0; */
      }
      return 1;
    }
    if (maxLines && !(lineCnt % maxLines) && !NextOFile(1, 1, preformatted))
      return 0;
  }
  return 1;
}

int Htmlize(htmlTupleSet& tuples, bool segmented, bool preformatted) {
  int bookmarkLine;
  bookmarkSet bookmarks;

  for (htmlTupleSet::iterator iter = tuples.begin();
       iter != tuples.end(); ++iter) {
    if ((*iter).htmlType == HtmlRsfTuple::Bookmark) {
      bookmarks.insert(*iter);
    }
  }

  for (htmlTupleSet::iterator iter = tuples.begin();
       iter != tuples.end(); ++iter) {
    HtmlRsfTuple& rsfTuple = *iter;

    /* Skip to next tuple if there is no anchor or bookmark */
    if (rsfTuple.htmlType == HtmlRsfTuple::NoType)
      continue;

    if (ifp == NULL) {
      /* Open the first input and output files */
      bookmarkLine = -1;
      if (!NextIFile(rsfTuple.rsf[3]) || !NextOFile(1, 0, preformatted))
	return 0;
    } else if (strcmp(iFile, rsfTuple.rsf[3])) {
      /* Copy remainder of input file is there is a file name change.
	 Then open new input and output files. */
      if (!CopyToLine(EOF, preformatted))
	return 0;
      bookmarkLine = -1;
      if (!NextIFile(rsfTuple.rsf[3]) || !NextOFile(0, 0, preformatted))
	return 0;
    }

    /* Copy to the referenced line number */
    if (!CopyToLine(rsfTuple.line, preformatted))
      return 0;

    if (rsfTuple.htmlType == HtmlRsfTuple::Bookmark) {

      /* Segment file if the first element is a bookmark.
	 If the file is segmented, there is no need for a bookmark. */
      if (segmented) {
	if (!NextOFile(1, 1, preformatted)) return 0;
      } else if (rsfTuple.line != bookmarkLine) {
	/* Only permit a single bookmark per line */
	bookmarkLine = rsfTuple.line;
	HtInsertBookmark(rec, bookmarkLine);
      }
    } else if (rsfTuple.htmlType == HtmlRsfTuple::Anchor) {
      /* read line, insert anchor(s) */
      bookmarkSet::iterator iter2;
      HtmlRsfTuple key(UniqueString::emptyString, UniqueString::emptyString, rsfTuple.rsf[2]);
      iter2 = bookmarks.find(key);
      if (iter2 != bookmarks.end()) {
	HtmlRsfTuple& bookmarkTuple = *iter2;
	if (bookmarkTuple.oFile.empty()) {
	  /* The bookmark reference may not be complete yet.
	     If it is missing, compose it. */
	  if (segmented) {
	    bookmarkTuple.oLine = 1;
	    MkOFileName(bookmarkTuple.rsf[3], tmpFile0, bookmarkTuple.line);
	  } else {
	    bookmarkTuple.oLine = bookmarkTuple.line;
	    MkOFileName(bookmarkTuple.rsf[3], tmpFile0, 1);
	  }
	  bookmarkTuple.oFile = tmpFile0;
	}
	HtInsertAnchor(rec, rsfTuple.rsf[2], bookmarkTuple.oFile,
		       bookmarkTuple.oLine);
      }
    }

    /* Save the location information for writing out RSF
       if it has not already been composed. */
    if (rsfTuple.oFile == UniqueString::emptyString) {
      rsfTuple.oFile = oFile;
      rsfTuple.oLine = segmented ? 1 : rsfTuple.line;
    }
  }
  if (!CopyToLine(EOF, preformatted))
    return 0;
  if (ifp != NULL)
    fclose(ifp);
  if (ofp != NULL) {
    char buff[MAXRECSIZE + 1];
    HtGetFileFtr(buff, NULL, preformatted);
    if (!WriteLine(buff, preformatted))
      return 0;
    fclose(ofp);
  }
  return 1;
}
